from typing import Any
from typing import Dict

import pop.contract


def client(hub, path: str, context):
    """
    Get an Vault client based on the path and call an operation on it
    idem will automatically populate credentials from acct in the client.

    The calls mirror the namespacing of hvac.Client and have the same parameters

    path::

        hvac.client.[service_path].[operation] [kwargs="values"]

    Examples:
        In these examples will use the service_path "secrets.kv.v2" and operation of "read_secret_version"

        Call from the CLI
        .. code-block: bash

            $ idem exec hvac.client.secrets.kv.v2.read_secret_version path="secret/test"

        Call from code
        .. code-block: python

            await hub.exec.hvac.client.secrets.kv.v2.read_secret_version(ctx, path="secret/test")

    :param hub:
    :param path: client.[service_path].[function_name]
    :param context: None
    :return: The result of the call
    """
    path_list = path.split(".")
    assert len(path_list) > 2
    c = path_list[0]
    service_path = path_list[1:-1]
    operation = path_list[-1]
    assert c == "client"

    async def _client_caller(ctx, *args, **kwargs) -> Dict[str, Any]:
        result = {"comment": (), "ret": None, "result": True}
        try:
            ret: Dict[str, Any] = await hub.tool.hvac.client.exec(
                ctx, service_path, operation, *args, **kwargs
            )
            if hasattr(ret, "keys"):
                keys = sorted(ret.keys())
            else:
                keys = []
            result["comment"] = tuple(keys)
            result["ret"] = ret
            result["result"] = bool(ret)
        except Exception as e:
            result["result"] = False
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
        return result

    return pop.contract.ContractedAsync(
        hub,
        contracts=[],
        func=_client_caller,
        ref=path,
        name=operation,
        implicit_hub=False,
    )



from typing import Any
from typing import Dict
from typing import List

import hvac

ITERATION_FINISHED = object()

__func_alias__ = {"exec_": "exec"}


async def exec_(
    hub,
    ctx,
    service_path: List[str],
    operation: str,
    *op_args: List[Any],
    **op_kwargs: Dict[str, Any],
) -> Any:
    """
    :param hub:
    :param ctx:
    :param service_path: The name of the service client to create
    :param operation: The operation to run from the service client
    :param op_args: arguments to pass to the operation call
    :param op_kwargs: keyword arguments to pass to the operation call

    :return: The result of the operation call
    """
    client = hvac.Client(url=ctx.acct.get("address"), token=ctx.acct.get("token"))
    # Don't pass kwargs that have a "None" value to the function call
    kwargs = {k: v for k, v in op_kwargs.items() if v is not None}
    hub.log.debug(f"Getting raw results for {'.'.join(service_path)}.{operation}")
    op = getattr(client, service_path[0])
    for step in service_path[1:]:
        op = getattr(op, step)
    op = getattr(op, operation)
    return await hub.pop.loop.wrap(op, *op_args, **kwargs)



import copy
from typing import Dict

import dict_tools.data

TREQ = {
    "unique": ["present", "absent"],
}

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    *,
    resource_id: (str, "alias=path"),
    key: str,
    value: str,
) -> Dict:
    """
    Creates or updates a secret key stored with Vault KV_v1 secret engine.

    Args:
        name (str):
            An Idem name of the resource.
        resource_id (str):
            The full logical path to write the data.
            This argument can also be specified using the alias "path." This should be prefixed with 'secret/'.
        key (str):
            Key of the secret
        value(str, Optional):
            Value of the secret

    Request Syntax:
        [vault-secret-name]:
          vault.secrets.kv_v1.key.present:
          - resource_id: 'string' # Can also be specified as "path"
          - key: 'string'
          - value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-secret:
              vault.secrets.kv_v1.key.present:
                - resource_id: secret/test # Can also be specified as "path"
                - key: my-birthday
                - value: 2012-10-17
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": (),
    }
    read_ret = await hub.exec.vault.secrets.kv_v1.secret.get(ctx, path=resource_id)
    data_to_update = {}
    if not read_ret["result"]:
        if "InvalidPath" not in str(read_ret["comment"]):
            result["result"] = False
            result["comment"] = read_ret["comment"]
            return result
    else:
        data_to_update = dict_tools.data.SafeNamespaceDict(read_ret["ret"]["data"])
        if data_to_update.get(key):
            result["old_state"] = {
                "name": name,
                "path": resource_id,
                "key": key,
                "value": data_to_update.get(key),
            }

    action = "create" if data_to_update.get(key) is None else "update"
    if (
        result["old_state"] is not None
        and result["old_state"]["key"] == key
        and result["old_state"]["value"]
        and result["old_state"]["value"] == value
    ):
        result["comment"] = (
            f"vault.secrets.kv_v1.key '{name}' nothing to be updated.",
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    if ctx.get("test", False):
        result["new_state"] = {
            "name": name,
            "path": resource_id,
            "key": key,
            "value": value,
        }
        result["comment"] = (f"Would {action} vault.secrets.kv_v1.key '{key}'.",)
        # calculating changes here to show keys that would be updated
        result["changes"] = hub.tool.vault.secret.calculate_changes(
            old_state={"data": result["old_state"]},
            new_state={"data": result["new_state"]},
        )
        return result

    data_to_update[key] = value
    write_ret = await hub.exec.hvac.client.secrets.kv.v1.create_or_update_secret(
        ctx, path=resource_id, secret=data_to_update
    )
    if not write_ret["result"]:
        result["result"] = False
        result["comment"] = write_ret["comment"]
        return result
    result["new_state"] = {
        "name": name,
        "path": resource_id,
        "key": key,
        "value": value,
    }
    # calculating changes here to show keys that would be updated
    result["changes"] = hub.tool.vault.secret.calculate_changes(
        old_state={"data": result["old_state"]}, new_state={"data": result["new_state"]}
    )
    if result["old_state"] is None or "create" == action:
        result["comment"] = (f"Created vault.secrets.kv_v1.key '{key}'.",)
    else:
        result["comment"] = (f"Updated vault.secrets.kv_v1.key '{key}'.",)
    return result


async def absent(
    hub,
    ctx,
    name: str,
    *,
    resource_id: (str, "alias=path"),
    key: str,
) -> Dict:
    """
    Deletes a secret key stored with Vault KV_v1 secret engine.

    Args:
        name (str):
            An Idem name of the resource.
        resource_id (str):
            The full logical path to write the data.
            This argument can also be specified using the alias "path." This should be prefixed with 'secret/'.
        key (str):
            Secret key to the deleted

    Request Syntax:
        [vault-secret-name]:
          vault.secrets.kv_v1.key.absent:
          - resource_id: 'string' # Can also be specified as "path"
          - key: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            my-secret:
              vault.secrets.kv_v1.key.absent:
                - resource_id: secret/test # Can also be specified as "path"
                - key: my-birthday
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": (),
    }
    current_data = {}
    read_ret = await hub.exec.vault.secrets.kv_v1.secret.get(ctx, path=resource_id)

    if not read_ret["result"]:
        if "InvalidPath" in str(read_ret["comment"]):
            result["comment"] = (f"vault.secrets.kv_v1.key '{key}' is already absent.",)
        else:
            result["result"] = False
            result["comment"] = read_ret["comment"]
        return result

    current_data = dict_tools.data.SafeNamespaceDict(read_ret["ret"]["data"])
    if key not in current_data:
        result["comment"] = (f"vault.secrets.kv_v1.key '{key}' is already absent.",)
        return result

    # "data" is not populated to reduce data exposure.
    result["old_state"] = {"name": name, "path": resource_id}

    if key not in current_data:
        result["comment"] = (f"vault.secrets.kv_v1.key '{key}' is already absent.",)
        return result

    if ctx.get("test", False):
        result["comment"] = (f"Would delete vault.secrets.kv_v1.key '{key}'.",)
        return result

    del current_data[key]
    if not current_data:
        # Delete path if all keys are removed. Path cannot exist with no keys
        ret = await hub.exec.hvac.client.secrets.kv.v1.delete_secret(
            ctx, path=resource_id
        )
    else:
        ret = await hub.exec.hvac.client.secrets.kv.v1.create_or_update_secret(
            ctx, path=resource_id, secret=current_data
        )

    if not ret["result"]:
        result["result"] = False
        result["comment"] = read_ret["comment"]
    else:
        result["comment"] = (f"Deleted vault.secrets.kv_v1.key '{key}'.",)
    return result


async def describe(hub, ctx):
    """
    Vault doesn't allow enumeration of keys
    """
    return {}



import json
from typing import ByteString
from typing import Tuple

__virtualname__ = "vault"


def __init__(hub):
    hub.source.vault.ACCT = ["vault"]


async def cache(
    hub, ctx, protocol: str, source: str, location: str
) -> Tuple[str, ByteString]:
    # Location ends with .sls
    if location.endswith(".sls"):
        location = location[:-4]

    if location != source:
        # This source wasn't specified with --params
        return "", b""

    version = ctx.acct.get("version", "v2")
    if version == "v1":
        data: bytes = await hub.exec.hvac.client.secrets.kv.v1.read_secret(
            ctx=ctx, path=location
        )
        return_data = (
            json.dumps(data["ret"].get("data")).encode("utf-8")
            if data["result"] and data["ret"]
            else None
        )
        return "", return_data
    else:
        data: bytes = await hub.exec.hvac.client.secrets.kv.v2.read_secret_version(
            ctx=ctx, path=location
        )
        return_data = (
            json.dumps(data["ret"].get("data").get("data")).encode("utf-8")
            if data["result"] and data["ret"] and data["ret"].get("data")
            else None
        )
        return "", return_data



"""
This plugin provides an interface for interacting with the raw hvac API using Idem's acct system and CLI paradigm.

All underlying calls are done asynchronously in a Thread-pool Executor
"""


def __init__(hub):
    # Provides the ctx argument to all execution modules
    # which will have profile info from the account module
    hub.exec.hvac.ACCT = ["vault"]

    # Load dynamic subs for accessing hvac client
    hub.pop.sub.dynamic(
        sub=hub.exec.hvac,
        subname="client",
        resolver=hub.tool.hvac.resolve.client,
        context=None,
    )
